/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/instruments/asianoption.hpp>
#include <ql/instruments/assetswap.hpp>
#include <ql/instruments/averagetype.hpp>
#include <ql/instruments/barrieroption.hpp>
#include <ql/instruments/barriertype.hpp>
#include <ql/instruments/basketoption.hpp>
#include <ql/instruments/bmaswap.hpp>
#include <ql/instruments/bond.hpp>
#include <ql/instruments/bondforward.hpp>
#include <ql/instruments/callabilityschedule.hpp>
#include <ql/instruments/capfloor.hpp>
#include <ql/instruments/claim.hpp>
#include <ql/instruments/cliquetoption.hpp>
#include <ql/instruments/compositeinstrument.hpp>
#include <ql/instruments/cpiswap.hpp>
#include <ql/instruments/cpicapfloor.hpp>
#include <ql/instruments/creditdefaultswap.hpp>
#include <ql/instruments/dividendbarrieroption.hpp>
#include <ql/instruments/dividendschedule.hpp>
#include <ql/instruments/dividendvanillaoption.hpp>
#include <ql/instruments/europeanoption.hpp>
#include <ql/instruments/fixedratebondforward.hpp>
#include <ql/instruments/floatfloatswap.hpp>
#include <ql/instruments/floatfloatswaption.hpp>
#include <ql/instruments/forward.hpp>
#include <ql/instruments/forwardrateagreement.hpp>
#include <ql/instruments/forwardvanillaoption.hpp>
#include <ql/instruments/futures.hpp>
#include <ql/instruments/impliedvolatility.hpp>
#include <ql/instruments/inflationcapfloor.hpp>
#include <ql/instruments/lookbackoption.hpp>
#include <ql/instruments/makecapfloor.hpp>
#include <ql/instruments/makecds.hpp>
#include <ql/instruments/makecms.hpp>
#include <ql/instruments/makeois.hpp>
#include <ql/instruments/makeswaption.hpp>
#include <ql/instruments/makevanillaswap.hpp>
#include <ql/instruments/makeyoyinflationcapfloor.hpp>
#include <ql/instruments/multiassetoption.hpp>
#include <ql/instruments/nonstandardswap.hpp>
#include <ql/instruments/nonstandardswaption.hpp>
#include <ql/instruments/oneassetoption.hpp>
#include <ql/instruments/overnightindexedswap.hpp>
#include <ql/instruments/overnightindexfuture.hpp>
#include <ql/instruments/payoffs.hpp>
#include <ql/instruments/quantobarrieroption.hpp>
#include <ql/instruments/quantoforwardvanillaoption.hpp>
#include <ql/instruments/quantovanillaoption.hpp>
#include <ql/instruments/stickyratchet.hpp>
#include <ql/instruments/stock.hpp>
#include <ql/instruments/swap.hpp>
#include <ql/instruments/swaption.hpp>
#include <ql/instruments/vanillaoption.hpp>
#include <ql/instruments/vanillastorageoption.hpp>
#include <ql/instruments/vanillaswingoption.hpp>
#include <ql/instruments/vanillaswap.hpp>
#include <ql/instruments/varianceswap.hpp>
#include <ql/instruments/yearonyearinflationswap.hpp>
#include <ql/instruments/zerocouponinflationswap.hpp>
#include <ql/instruments/zerocouponswap.hpp>

#include <ql/instruments/bonds/all.hpp>
