/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/math/abcdmathfunction.hpp>
#include <ql/math/array.hpp>
#include <ql/math/autocovariance.hpp>
#include <ql/math/bernsteinpolynomial.hpp>
#include <ql/math/beta.hpp>
#include <ql/math/bspline.hpp>
#include <ql/math/comparison.hpp>
#include <ql/math/curve.hpp>
#include <ql/math/errorfunction.hpp>
#include <ql/math/factorial.hpp>
#include <ql/math/fastfouriertransform.hpp>
#include <ql/math/functional.hpp>
#include <ql/math/generallinearleastsquares.hpp>
#include <ql/math/kernelfunctions.hpp>
#include <ql/math/incompletegamma.hpp>
#include <ql/math/interpolation.hpp>
#include <ql/math/lexicographicalview.hpp>
#include <ql/math/linearleastsquaresregression.hpp>
#include <ql/math/matrix.hpp>
#include <ql/math/modifiedbessel.hpp>
#include <ql/math/pascaltriangle.hpp>
#include <ql/math/polynomialmathfunction.hpp>
#include <ql/math/primenumbers.hpp>
#include <ql/math/quadratic.hpp>
#include <ql/math/rounding.hpp>
#include <ql/math/richardsonextrapolation.hpp>
#include <ql/math/sampledcurve.hpp>
#include <ql/math/solver1d.hpp>
#include <ql/math/transformedgrid.hpp>

#include <ql/math/copulas/all.hpp>
#include <ql/math/distributions/all.hpp>
#include <ql/math/integrals/all.hpp>
#include <ql/math/interpolations/all.hpp>
#include <ql/math/matrixutilities/all.hpp>
#include <ql/math/ode/all.hpp>
#include <ql/math/optimization/all.hpp>
#include <ql/math/randomnumbers/all.hpp>
#include <ql/math/solvers1d/all.hpp>
#include <ql/math/statistics/all.hpp>
