/* This file is automatically generated; do not edit.     */
/* Add the files to be included into Makefile.am instead. */

#include <ql/math/copulas/alimikhailhaqcopula.hpp>
#include <ql/math/copulas/claytoncopula.hpp>
#include <ql/math/copulas/farliegumbelmorgensterncopula.hpp>
#include <ql/math/copulas/frankcopula.hpp>
#include <ql/math/copulas/galamboscopula.hpp>
#include <ql/math/copulas/gaussiancopula.hpp>
#include <ql/math/copulas/gumbelcopula.hpp>
#include <ql/math/copulas/huslerreisscopula.hpp>
#include <ql/math/copulas/independentcopula.hpp>
#include <ql/math/copulas/marshallolkincopula.hpp>
#include <ql/math/copulas/maxcopula.hpp>
#include <ql/math/copulas/mincopula.hpp>
#include <ql/math/copulas/plackettcopula.hpp>

